#include "sys/types.h"
#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/cred.h"
#include "sys/ddi.h"
#include "sys/inline.h"
#include "sys/systm.h"

#include "sys/cmn_err.h"

#include "sys/midi.h"

static struct module_info rminfo = { 0xaabb, "midi", 0, INFPSZ, 0, 0 };
static struct module_info wminfo = { 0xaabb, "midi", 0, INFPSZ, 0, 0 };

static int midiopen(), midiclose(), midiwput(), midireset();
static void midiout(), midiioc(), midisetlast();

static struct qinit rinit = {
	NULL, NULL, midiopen, midiclose, NULL, &rminfo, NULL };

static struct qinit winit = {
	midiwput, NULL, NULL, NULL, NULL, &wminfo, NULL };

struct streamtab midiinfo = { &rinit, &winit, NULL, NULL };

static int mididebug = 0;

#define DATA_PORT 0x330
#define STATUS_PORT 0x331
#define DATA_READY 0x40
#define DATA_AVAILABLE 0x80
#define RESET 0xff
#define UART 0x3f
/* note that active sensing and ACK from controller are the same */
#define ACK 0xfe
#define ACTIVE 0xfe

#define midi_putdata(ctlr,c)	(outb(ctlr_to_data_port[ctlr],(c)))
#define midi_putcmd(ctlr,c)	(outb(ctlr_to_status_port[ctlr],(c)))
#define midi_getstatus(ctlr)	(inb(ctlr_to_status_port[ctlr]))
#define ready_for_data(ctlr)	(!(midi_getstatus(ctlr)&DATA_READY))
#define midi_getdata(ctlr)	(inb(ctlr_to_data_port[ctlr]))
#define data_is_available(ctlr)	(!(midi_getstatus(ctlr)&DATA_AVAILABLE))
#define data_is_not_available(ctlr)	(midi_getstatus(ctlr)&DATA_AVAILABLE)

extern int mididevflag = 0;

static int
midiopen(q, devp, flag, sflag, credp)
queue_t *q;	/* read queue */
dev_t *devp;	/* major/minor device number */
int flag;	/* file flags */
int sflag;	/* stream open flags */
cred_t *credp;	/* credentials */
{
	struct midi *midi;
	int ctlr, n;

#ifdef lint
	flag = flag;
#endif
	if ( mididebug )
		cmn_err(CE_NOTE,"midiopen start flag=%d sflag=%d *devp=0x%x q=0x%x credp=0x%x",flag,sflag,(int)(*devp),(int)q,(int)credp);

	if ( sflag )	/* check if non-driver open (or CLONEOPEN) */
		return ENXIO;

	ctlr = getminor(*devp);
	if ( ctlr >= midi_nctlr )
		return ENXIO;

	/* all opens are cloning, look for unused entry */
	for ( midi=midi_midi,n=0; midi<midi_end; midi++,n++ ) {
		if ( (midi->flags & ISOPEN) == 0 )
			break;
	}
	if ( midi >= midi_end )
		return EAGAIN; /* couldn't find unused structure */

	*devp = makedevice(getmajor(*devp), 0x10 | n);
	if ( mididebug )
		cmn_err(CE_NOTE,"midiopen using stream n=%d new*devp=0x%x",n,*devp);

	midi->rq = q;
	midi->wq = WR(q);
	q->q_ptr = (char*) midi;
	WR(q)->q_ptr = (char*) midi;

	if ( ((midi->flags)&ISOPEN) != 0 )
		cmn_err(CE_NOTE,"midi stream already open!?");

	midi->flags |= ISOPEN;

	if ( midireset(ctlr,1) ) {	 /* default is UART mode */
		midi->flags &= (~ISOPEN);
		return EIO;
	}

	midi->flags &= (~ACTSENSE); /* default is to ignore active sensing */
	midi->clockoffset = lbolt;  /* TIME starts out at 0 */
	midi->ctlr = ctlr;

	midisetlast();
	return 0;	/* success */
}

static void
midisetlast()
{
	register struct midi *m, *lm;

	lm = midi_end - 1;
	for ( m=lm; m>=midi_midi; m-- ) {
		if ( m->flags & ISOPEN ) {
			lm = m;
			break;
		}
	}
	midi_lastopen = lm + 1;
}

static void
midiack(q,mp)
queue_t *q;	/* pointer to the queue */
mblk_t *mp;	/* message pointer */
{
	mp->b_datap->db_type = M_IOCACK;
	((struct iocblk *)(mp->b_rptr))->ioc_error = 0;
	qreply(q,mp);
}

static void
midinak(q,mp)
queue_t *q;	/* pointer to the queue */
mblk_t *mp;	/* message pointer */
{
	mp->b_datap->db_type = M_IOCNAK;
	((struct iocblk *)(mp->b_rptr))->ioc_error = EINVAL;
	qreply(q,mp);
}

static int
midiwput(q, mp)
queue_t *q;	/* pointer to the queue */
mblk_t *mp;	/* message pointer */
{
	struct copyresp *csp;
	struct iocblk *iocbp;

	switch (mp->b_datap->db_type) {
	case M_DATA:
		midiout((struct midi *)q->q_ptr,
			mp->b_rptr,
			mp->b_wptr - mp->b_rptr);
		freemsg(mp);
		break;
	case M_FLUSH:
		if ( *mp->b_rptr & FLUSHW ) {
			flushq(q,FLUSHDATA);
		}
		if ( *mp->b_rptr & FLUSHR ) {
			*mp->b_rptr &= ~FLUSHW;
			qreply(q,mp);
		}
		else {
			freemsg(mp);
		}
		break;
	case M_IOCDATA:
		/* probably a response to MIDISTAT */
		csp = (struct copyresp *)mp->b_rptr;
		if ( csp->cp_cmd != MIDISTAT && csp->cp_cmd != MIDITIME ) {
			/* if this was a module, mp would be passed on */
			cmn_err(CE_NOTE,"M_IOCDATA not MIDISTAT?");
			freemsg(mp);
			break;
		}
		if ( csp->cp_rval ) {
			cmn_err(CE_NOTE,"M_IOCDATA  cp_rval != 0 ?");
			freemsg(mp);	/* failure */
			break;
		}
		mp->b_datap->db_type = M_IOCACK;
		mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
		iocbp = (struct iocblk *)(mp->b_rptr);
		iocbp->ioc_error = 0;
		iocbp->ioc_count = 0;
		iocbp->ioc_rval = 0;
		qreply(q,mp);
		break;
	case M_IOCTL:
		midiioc(q,mp);
		break;
	default:
		freemsg(mp);
		break;
	}
	return 0;
}

void
midiioc(q,mp)
queue_t *q;
mblk_t *mp;
{
	register struct midi *midi;
	int transparent;
	struct iocblk *iocbp;
	struct copyreq *cqp;
	struct midi_stat *sp;

	midi = (struct midi *)q->q_ptr;
	iocbp = (struct iocblk *)(mp->b_rptr);
	switch (iocbp->ioc_cmd) {
	case MIDIACTIVE:
		midi->flags |= ACTSENSE;
		midiack(q,mp);
		break;
	case MIDINOACTIVE:
		midi->flags &= (~ACTSENSE);
		midiack(q,mp);
		break;
	case MIDIRESET:
		if ( midireset(midi->ctlr,1) )
			midinak(q,mp);
		else
			midiack(q,mp);
		break;
	case MIDITHRU:
		if ( midireset(midi->ctlr,0) )
			midinak(q,mp);
		else
			midiack(q,mp);
		break;
	case MIDITIMERESET:
		midi->clockoffset = lbolt;
		midiack(q,mp);
		break;
	case MIDISTAT:
		if ( iocbp->ioc_count == TRANSPARENT ) {
			transparent = 1;
			cqp = (struct copyreq *)mp->b_rptr;
			cqp->cq_size = sizeof(struct midi_stat);
			cqp->cq_addr = (caddr_t) *(long*)(mp->b_cont->b_rptr);
			cqp->cq_flag = 0;
		}
		/* Whether it's an I_STR ioctl or a TRANSPARENT one, we */
		/* put the data to be returned into the b_cont data block */
		if ( mp->b_cont )
			freemsg(mp->b_cont);
		mp->b_cont = allocb(sizeof(struct midi_stat),0);
		if ( mp->b_cont==NULL || (sp=(struct midi_stat*)(mp->b_cont->b_rptr))==NULL ) {
			cmn_err(CE_NOTE,"MIDISTAT failed cont or sp==NULL");
			mp->b_datap->db_type = M_IOCNAK;
			iocbp->ioc_error = EAGAIN;
			qreply(q,mp);
			break;
		}
		mp->b_cont->b_wptr += sizeof(struct midi_stat);
		sp->clock = ((1000/HZ)*(lbolt - midi->clockoffset));

		if ( midi->rq!=NULL && midi->rq->q_next != NULL )
			sp->rqcount = midi->rq->q_next->q_count;
		else
			sp->rqcount = 0;

		if ( transparent ) {
			mp->b_datap->db_type = M_COPYOUT;
			mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
			cqp = (struct copyreq *)mp->b_rptr;
		}
		else {
			mp->b_datap->db_type = M_IOCACK;
			iocbp->ioc_count = sizeof(struct midi_stat);
		}
		qreply(q,mp);
		break;

/* The MIDITIME ioctl is going to be phased out in favor of MIDISTAT. */
	case MIDITIME:
		if ( iocbp->ioc_count == TRANSPARENT ) {
			transparent = 1;
			cqp = (struct copyreq *)mp->b_rptr;
			cqp->cq_size = sizeof(long);
			cqp->cq_addr = (caddr_t) *(long*)(mp->b_cont->b_rptr);
			cqp->cq_flag = 0;
		}
		/* Whether it's an I_STR ioctl or a TRANSPARENT one, we */
		/* put the data to be returned into the b_cont data block */
		if ( mp->b_cont )
			freemsg(mp->b_cont);
		mp->b_cont = allocb(sizeof(long),0);
		{ long *lp;
		if ( mp->b_cont==NULL || (lp=(long*)(mp->b_cont->b_rptr))==NULL ) {
			cmn_err(CE_NOTE,"MIDISTAT failed cont or sp==NULL");
			mp->b_datap->db_type = M_IOCNAK;
			iocbp->ioc_error = EAGAIN;
			qreply(q,mp);
			break;
		}
		mp->b_cont->b_wptr += sizeof(long);
		*lp = ((1000/HZ)*(lbolt - midi->clockoffset));
		}

		if ( transparent ) {
			mp->b_datap->db_type = M_COPYOUT;
			mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
			cqp = (struct copyreq *)mp->b_rptr;
		}
		else {
			mp->b_datap->db_type = M_IOCACK;
			iocbp->ioc_count = sizeof(long);
		}
		qreply(q,mp);
		break;

	default:
		midinak(q,mp);
		break;
	}
}

static int
midiclose(q,flag,credp)
queue_t *q;	/* read queue */
int flag;	/* file flags */
cred_t *credp;	/* credentials */
{
	struct midi *midi, *m;
	int n;

#ifdef lint
	flag = flag;
	credp = credp;
#endif
	midi = (struct midi *)q->q_ptr;
	if ( midi == NULL ) {
		cmn_err(CE_NOTE,"Hey, midi==NULL in midiclose()!");
		return 1;
	}

	/* See if any other streams have that MPU controller open. */
	for ( m=midi_midi; m<midi_end; m++ ) {
		if ( m!=midi && ((m->flags)&ISOPEN)!=0 && m->ctlr == midi->ctlr )
			break;
	}
	/* only reset if noone else has it open */
	if ( m >= midi_end )
		(void) midireset(midi->ctlr,0);

	if ( ((midi->flags)&ISOPEN) == 0 ) {
		cmn_err(CE_NOTE,"Hey, closing a non-open midi stream?!");
		return 0;
	}

	midi->flags &= (~ISOPEN);
	midisetlast();
	return 0;
}

void
midiintr(irq)
int irq;
{
	register struct midi *m;
	register int ctlr;
	register unsigned char c;
	mblk_t *mp;

	ctlr = irq_to_ctlr[irq];
	if ( data_is_not_available(ctlr) )
		return;
	c = midi_getdata(ctlr);
	for ( m=midi_midi; m<midi_lastopen; m++ ) {
		if ( ((m->flags)&ISOPEN) && m->ctlr == ctlr ) {
			/* ACK (e.g. in response to midireset()) and */
			/* ACTIVE are the same */
			if ( c == ACK ) {
				if ( (m->flags & WAITINGFORACK) != 0 ) {
					untimeout(m->to_id);
					m->flags &= (~WAITINGFORACK);
					wakeup((caddr_t)m);
				}
				/* active sensing is ignored */
			}
			else {
				mp = allocb(sizeof(long),0);
				if ( mp == NULL )
					return;
				*(mp->b_wptr++) = c;
				putnext(m->rq,mp);
			}
		}
	}
}

void
midiinit()
{
	register int n;

	/* create an irq_to_ctlr array so that midiintr() can */
	/* get the ctlr number quickly. */
	for ( n=0; n<NUMIRQ; n++ )
		irq_to_ctlr[n] = 0;
	for ( n=0; n<midi_nctlr; n++ ) {
		irq_to_ctlr[ ctlr_to_irq[n] ] = n;
	}
}

void
midistart()
{
	/* If you want to allocate any buffers or stuff, do it here */
}

/*******
 * From here on is the MPU-specific stuff.
 ******/

int
midireset(ctlr,uart)
{
	register struct midi *m = &midi_midi[ ctlr ];
	register time_t etime;
	int n, oldlev;
	int gotack = 0;

	while ( data_is_available(ctlr) )
		(void) midi_getdata(ctlr);

	/* try 5 times to get an ACK after a RESET */
	for ( n=0; n<5 && !gotack; n++ ) {
		midi_putcmd(ctlr,RESET);
		oldlev = splhi();
		m->flags |= WAITINGFORACK;
		/* timeout in 20 milliseconds */
		m->to_id = timeout(wakeup,(caddr_t)m,HZ/50);
		if ( sleep((caddr_t)m,(PCATCH | PZERO+2)) == 1 ) {
			/* if user sent a signal to abort it */
			m->flags &= (~WAITINGFORACK);
			untimeout(m->to_id);
			splx(oldlev);
			return 1;
		}
		if ( (m->flags & WAITINGFORACK) == 0 )
			gotack = 1;
		splx(oldlev);
	}
	if ( ! gotack )
		return 1;
	if ( uart )
		midi_putcmd(ctlr,UART);	/* doesn't send an ACK back */
	return 0;
}

static int
midi_wait_for_ready(ctlr)
{
	int tmout;

	for ( tmout=50000; tmout>0; tmout-- ) {
		if ( ready_for_data(ctlr) )
			break;
	}
	if ( tmout <= 0 ) {
		cmn_err(CE_NOTE,"/dev/midi timed out waiting for data ready\n");
		return 0;
	}
	return 1;
}

static void
midiout(m,base,count)
struct midi *m;
unsigned char *base;
unsigned int count;
{
	register int ctlr;

	if ( !m ) {
		cmn_err(CE_NOTE,"m==NULL in midiout!!");
		return;
	}
	ctlr = m->ctlr;
	while ( count ) {
		if ( midi_wait_for_ready(ctlr) ) {
			midi_putdata(ctlr,*base++);
			count--;
		}
	}
}
